import torch


class PlusResidual(torch.nn.Module):
    """
    Wraps a layer such that the forward pass returns
    ```x + self.layer(x)```
    """

    def __init__(self, layer: torch.nn.Module):
        """
        Arguments:
            layer: The layer to be wrapped.
        """
        super().__init__()
        self.layer = layer

    def forward(self, x):
        """"""
        return x + self.layer(x)



import torch
from torchmetrics.functional import accuracy

from ..utils import common_functions as c_f
from .base_validator import BaseValidator


class AccuracyValidator(BaseValidator):
    """
    Returns accuracy using the
    [torchmetrics accuracy function](https://torchmetrics.readthedocs.io/en/latest/references/functional.html#accuracy-func)
    """

    def __init__(self, layer="preds", torchmetric_kwargs=None, **kwargs):
        super().__init__(**kwargs)
        self.layer = layer
        self.torchmetric_kwargs = c_f.default(torchmetric_kwargs, {})
        self.accuracy_func = accuracy

    def compute_score(self, src_val):
        return self.accuracy_func(
            src_val[self.layer], src_val["labels"], **self.torchmetric_kwargs
        ).item()



from typing import Callable, List, Union

from ..utils import common_functions as c_f
from .base import BaseConditionHook, BaseHook, BaseWrapperHook


class EmptyHook(BaseHook):
    """Returns two empty dictionaries."""

    def call(self, losses, inputs):
        """"""
        return {}, {}

    def _loss_keys(self):
        """"""
        return []

    def _out_keys(self):
        """"""
        return []


class ZeroLossHook(BaseHook):
    """
    Returns only 0 losses and ```None``` outputs.
    """

    def __init__(self, loss_names: List[str], out_names: List[str], **kwargs):
        """
        Arguments:
            loss_names: The keys of the loss dictionary
                which will have ```tensor(0.)``` as its values.
            out_names: The keys of the output dictionary
                which will have ```None``` as its values.
        """
        super().__init__(**kwargs)
        self.loss_names = loss_names
        self.out_names = out_names

    def call(self, losses, inputs):
        """"""
        out_keys = set(self.out_names) - inputs.keys()
        return {k: c_f.zero_loss() for k in self.loss_names}, {
            k: None for k in out_keys
        }

    def _loss_keys(self):
        """"""
        return self.loss_names

    def _out_keys(self):
        """"""
        return self.out_names


class ChainHook(BaseHook):
    """
    Calls multiple hooks sequentially.
    The Nth hook receives the context accumulated through hooks 0 to N-1.
    """

    def __init__(
        self,
        *hooks: BaseHook,
        conditions: List[BaseConditionHook] = None,
        alts: List[BaseHook] = None,
        overwrite: Union[bool, List[int]] = False,
        **kwargs,
    ):
        """
        Arguments:
            hooks: a sequence of hooks that will be called sequentially.
            conditions: an optional list of condition hooks.
                If conditions[i] returns False, then alts[i] is called. Otherwise hooks[i] is called.
            alts: an optional list of hooks that will be executed
                when the corresponding condition hook returns False
            overwrite: If True, then hooks will be allowed to overwrite keys in the context.
                If a list of integers, then the hooks at the specified indices
                will be allowed to overwrite keys in the context.

        """

        super().__init__(**kwargs)
        self.hooks = hooks
        self.conditions = c_f.default(
            conditions, [TrueHook() for _ in range(len(hooks))]
        )
        self.alts = c_f.default(
            alts, [ZeroLossHook(h.loss_keys, h.out_keys) for h in self.hooks]
        )
        self.check_alt_keys_match_hook_keys()
        if not isinstance(overwrite, (list, bool)):
            raise TypeError("overwrite must be a list or bool")
        self.overwrite = overwrite
        self.in_keys = self.hooks[0].in_keys

    def call(self, losses, inputs):
        """"""
        out_losses, outputs = {}, {}
        all_losses, all_inputs = losses, inputs
        prev_losses, prev_outputs = {}, {}
        for i, h in enumerate(self.hooks):
            self.check_overwrite(i, all_losses, prev_losses, False)
            self.check_overwrite(i, all_inputs, prev_outputs, self.overwrite)
            all_losses = {**all_losses, **prev_losses}
            all_inputs = {**all_inputs, **prev_outputs}
            if self.conditions[i](all_losses, all_inputs):
                x = h(all_losses, all_inputs)
            else:
                x = self.alts[i](all_losses, all_inputs)
            prev_losses, prev_outputs = x
            out_losses.update(prev_losses)
            outputs.update(prev_outputs)
        return out_losses, outputs

    def check_overlap(self, x, y, names):
        is_overlap, overlap = c_f.dicts_are_overlapping(x, y, return_overlap=True)
        if is_overlap:
            raise KeyError(
                f"overwrite is false, but {names[0]} and {names[1]} have overlapping keys: {overlap}"
            )

    def check_overwrite(self, i, kwargs, prev_outputs, overwrite):
        if not overwrite or (isinstance(overwrite, list) and i not in overwrite):
            self.check_overlap(kwargs, prev_outputs, ["kwargs", "prev_outputs"])

    def _loss_keys(self):
        """"""
        return c_f.join_lists([h.loss_keys for h in self.hooks])

    def _out_keys(self):
        """"""
        return c_f.join_lists([h.out_keys for h in self.hooks])

    @property
    def last_hook_out_keys(self):
        return self.hooks[-1].out_keys

    def check_alt_keys_match_hook_keys(self):
        for i in range(len(self.hooks)):
            h = self.hooks[i]
            a = self.alts[i]
            if (sorted(h.loss_keys) != sorted(a.loss_keys)) or (
                sorted(h.out_keys) != sorted(a.out_keys)
            ):
                raise ValueError(
                    "alt loss/out keys must be equal to hook loss/out keys"
                )

    def children_repr(self):
        x = super().children_repr()
        x["hooks"] = self.hooks
        if any(not isinstance(c, TrueHook) for c in self.conditions):
            x.update({"conditions": self.conditions, "alts": self.alts})
        return x


# outputs are not shared among parallel streams
# they are returned at the very end
class ParallelHook(BaseHook):
    """
    Calls multiple hooks while keeping contexts separate.
    The Nth hook receives the same context as hooks 0 to N-1.
    All the output contexts are merged at the end.
    """

    def __init__(self, *hooks: BaseHook, **kwargs):
        """
        Arguments:
            hooks: a sequence of hooks that will be called sequentially,
                with each hook receiving the same initial context.
        """
        super().__init__(**kwargs)
        self.hooks = hooks
        self.in_keys = c_f.join_lists([h.in_keys for h in self.hooks])

    def call(self, losses, inputs):
        """"""
        out_losses, outputs = {}, {}
        for h in self.hooks:
            x = h(losses, inputs)
            out_losses.update(x[0])
            outputs.update(x[1])

        return out_losses, outputs

    def children_repr(self):
        x = super().children_repr()
        x.update({"hooks": self.hooks})
        return x

    def _loss_keys(self):
        """"""
        return c_f.join_lists([h.loss_keys for h in self.hooks])

    def _out_keys(self):
        """"""
        return c_f.join_lists([h.out_keys for h in self.hooks])


class OnlyNewOutputsHook(BaseWrapperHook):
    """
    Returns only outputs that are not present in the input context.
    You should use this if you want to change the value of
    a key passed to self.hook, but not propagate that change
    to the outside.
    """

    def __init__(self, hook: BaseHook, **kwargs):
        """
        Arguments:
            hook: The hook inside which changes to the context will be allowed.
        """
        super().__init__(**kwargs)
        self.hook = hook

    def call(self, losses, inputs):
        """"""
        losses, outputs = self.hook(losses, inputs)
        outputs = {k: outputs[k] for k in (outputs.keys() - inputs.keys())}
        c_f.assert_dicts_are_disjoint(inputs, outputs)
        return losses, outputs


class ApplyFnHook(BaseHook):
    """
    Applies a function to specific values of the context.
    """

    def __init__(
        self, fn: Callable, apply_to: List[str], is_loss: bool = False, **kwargs
    ):
        """
        Arguments:
            fn: The function that will be applied to the inputs.
            apply_to: fn will be applied to ```inputs[k]``` for k in apply_to
            is_loss: If False, then the returned loss dictionary will be empty.
                Otherwise, the returned output dictionary will be empty.
        """
        super().__init__(**kwargs)
        self.fn = fn
        self.apply_to = apply_to
        self.is_loss = is_loss

    def call(self, losses, inputs):
        """"""
        x = c_f.extract(inputs, self.apply_to)
        outputs = {k: self.fn(v) for k, v in zip(self.apply_to, x)}
        if self.is_loss:
            return outputs, {}
        return {}, outputs

    def _loss_keys(self):
        """"""
        return self.apply_to if self.is_loss else []

    def _out_keys(self):
        """"""
        return [] if self.is_loss else self.apply_to

    def extra_repr(self):
        return c_f.extra_repr(self, ["apply_to"])


class TrueHook(BaseConditionHook):
    """Returns ```True```"""

    def call(self, losses, inputs):
        """"""
        return True


class FalseHook(BaseConditionHook):
    """Returns ```False```"""

    def call(self, losses, inputs):
        """"""
        return False


class NotHook(BaseConditionHook):
    """Returns the boolean negation of the wrapped hook."""

    def __init__(self, hook: BaseConditionHook, **kwargs):
        """
        Arguments:
            hook: The condition hook that will be negated.
        """
        super().__init__(**kwargs)
        self.hook = hook

    def call(self, losses, inputs):
        """"""
        return not self.hook(losses, inputs)


class AssertHook(BaseWrapperHook):
    """
    Asserts that the output keys of a hook match a specified regex string
    """

    def __init__(self, hook: BaseHook, allowed: str, **kwargs):
        """
        Arguments:
            hook: The wrapped hook
            allowed: The output dictionary of ```hook```
                must have keys that match the ```allowed``` regex.
        """
        super().__init__(**kwargs)
        self.hook = hook
        if not isinstance(allowed, str):
            raise TypeError("allowed must be a str")
        self.allowed = allowed

    def call(self, losses, inputs):
        """"""
        losses, outputs = self.hook(losses, inputs)
        self.assert_fn(outputs)
        return losses, outputs

    def assert_fn(self, outputs):
        filtered = c_f.filter(outputs, self.allowed)
        if len(filtered) != len(outputs):
            error_str = f"{c_f.cls_name(self.hook)} is producing outputs that don't match the allowed regex in {c_f.cls_name(self)}\n"
            error_str += f"output keys = {outputs.keys()}\n"
            error_str += f"regex filter = {self.allowed}"
            raise ValueError(error_str)

    def extra_repr(self):
        return c_f.extra_repr(self, ["allowed"])


class MultiplierHook(BaseWrapperHook):
    """
    Multiplies every loss by a scalar
    """

    def __init__(self, hook: BaseHook, m: float, **kwargs):
        """
        Arguments:
            hook: The losses of this hook will be multiplied by ```m```
            m: The scalar
        """
        super().__init__(**kwargs)
        self.hook = hook
        self.m = m

    def call(self, losses, inputs):
        """"""
        losses, outputs = self.hook(losses, inputs)
        losses = {k: v * self.m for k, v in losses.items()}
        return losses, outputs

    def extra_repr(self):
        return c_f.extra_repr(self, ["m"])


class RepeatHook(BaseHook):
    """
    Executes the wrapped hook ```n``` times.
    """

    def __init__(self, hook: BaseHook, n: int, keep_only_last: bool = False, **kwargs):
        """
        Arguments:
            hook: The hook that will be executed ```n``` times
            n: The number of times the hook will be executed.
            keep_only_last: If ```False```, the (losses, outputs) from each execution
                will be accumulated, and the keys will have the iteration number appended.
                If ```True```, then only the (losses, outputs) of the final execution will
                be kept.
        """
        super().__init__(**kwargs)
        self.hook = hook
        self.n = n
        self.keep_only_last = keep_only_last

    def call(self, losses, inputs):
        """"""
        losses, outputs = {}, {}
        for i in range(self.n):
            x = self.hook(losses, inputs)
            if self.keep_only_last and i == self.n - 1:
                losses, outputs = x
            else:
                losses.update({f"{k}{i}": v for k, v in x[0].items()})
                outputs.update({f"{k}{i}": v for k, v in x[1].items()})
        return losses, outputs

    def _loss_keys(self):
        """"""
        if self.keep_only_last:
            return self.hook.loss_keys
        else:
            return [f"{k}{i}" for k in self.hook.loss_keys for i in range(self.n)]

    def _out_keys(self):
        """"""
        if self.keep_only_last:
            return self.hook.out_keys
        else:
            return [f"{k}{i}" for k in self.hook.out_keys for i in range(self.n)]

    def extra_repr(self):
        return c_f.extra_repr(self, ["n", "keep_only_last"])


class ApplyToListHook(BaseWrapperHook):
    def __init__(self, hook, list_size, regex, **kwargs):
        super().__init__(**kwargs)
        self.hook = hook
        self.list_size = list_size
        self.regex = regex

    def call(self, losses, inputs):
        """"""
        search_strs = c_f.filter(inputs, self.regex)
        extracted = c_f.extract(inputs, search_strs)
        if len(set(len(x) for x in extracted)) != 1:
            raise TypeError(f"All extracted input lists must be equal in length")
        losses = {}
        for i in range(len(extracted[0])):
            curr_extracted = [x[i] for x in extracted]
            curr_extracted = {k: v for k, v in zip(search_strs, curr_extracted)}
            x = self.hook(losses, {**inputs, **curr_extracted})
            losses.update({f"{k}{i}": v for k, v in x[0].items()})
        return losses, {}

    def _loss_keys(self):
        """"""
        return [f"{k}{i}" for k in self.hook.loss_keys for i in range(self.list_size)]

    def _out_keys(self):
        """"""
        return []

    def extra_repr(self):
        return c_f.extra_repr(self, ["list_size", "regex"])



import copy
from typing import Tuple, Union

import torch

from ..utils import common_functions as c_f


class ModelWithBridge(torch.nn.Module):
    """
    Implementation of the bridge architecture described in
    [Gradually Vanishing Bridge for Adversarial Domain Adaptation](https://arxiv.org/abs/2003.13183).
    """

    def __init__(self, model: torch.nn.Module, bridge: torch.nn.Module = None):
        """
        Arguments:
            model: Any pytorch model.
            bridge: A model which has the same input/output sizes as ```model```.
                If ```None```, then the bridge is formed by copying ```model```,
                and randomly reinitialization all its parameters.
        """
        super().__init__()
        self.model = model
        if bridge is None:
            bridge = c_f.reinit(copy.deepcopy(model))
        self.bridge = bridge

    def forward(
        self, x: torch.Tensor, return_bridge: bool = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Arguments:
            x: The input to both ```self.model``` and ```self.bridge```.
            return_bridge: Whether or not to return the bridge output
                in addition to the ```model - bridge``` output
        Returns:
            If ```return_bridge = False```, then return just ```model - bridge```.
            If ```return_bridge = True```, then return a tuple of ```(model - bridge), bridge```
        """
        y = self.model(x)
        z = self.bridge(x)
        output = y - z
        if return_bridge:
            return output, z
        return output



from .base_weighter import BaseWeighter, weight_losses


def mean(losses):
    return sum(losses) / len(losses)


def mean_weighter(kwargs):
    return weight_losses(mean, {}, 1, kwargs)


class MeanWeighter(BaseWeighter):
    """
    Weights the losses and then returns the **mean** of the weighted losses.
    """

    def __init__(self, **kwargs):
        super().__init__(reduction=mean, **kwargs)



from .diversity_validator import DiversityValidator
from .entropy_validator import EntropyValidator
from .multiple_validators import MultipleValidators


class EntropyDiversityValidator(MultipleValidators):
    def __init__(self, **kwargs):
        validators = {"entropy": EntropyValidator(), "diversity": DiversityValidator()}
        super().__init__(validators=validators, **kwargs)



from ..utils import common_functions as c_f
from .conditions import StrongDHook
from .features import FeaturesHook, FeaturesWithGradAndDetachedHook, FrozenModelHook
from .gan import GANHook
from .utils import ChainHook, EmptyHook, FalseHook, TrueHook


class ADDAHook(GANHook):
    """
    Implementation of
    [Adversarial Discriminative Domain Adaptation](https://arxiv.org/abs/1702.05464).

    Extends [```GANHook```][pytorch_adapt.hooks.gan.GANHook].
    """

    def __init__(self, threshold: float = 0.6, pre_g=None, post_g=None, **kwargs):
        """
        Arguments:
            threshold: In each training iteration, the generator is only updated
                if the discriminator's accuracy is greater than ```threshold```.
        """
        [pre_g, post_g] = c_f.many_default([pre_g, post_g], [[], []])
        sf_frozen = FrozenModelHook(FeaturesHook(detach=True, domains=["src"]), "G")
        tf_all = FeaturesWithGradAndDetachedHook(model_name="T", domains=["target"])
        pre_d = ChainHook(sf_frozen, tf_all)
        num_pre_g = len(pre_g)
        gen_conditions = [TrueHook() for _ in range(num_pre_g + len(post_g) + 2)]
        # generator condition, classifier condition
        gen_conditions[num_pre_g : num_pre_g + 2] = [
            StrongDHook(threshold),
            FalseHook(),
        ]
        super().__init__(
            pre_d=[pre_d],
            pre_g=pre_g,
            post_g=post_g,
            gen_conditions=gen_conditions,
            gen_domains=["target"],
            c_hook=EmptyHook(),
            **kwargs
        )



import re
from abc import ABC, abstractmethod
from contextlib import nullcontext
from typing import Callable, List

import torch

from ..layers import EntropyWeights
from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .features import FeaturesAndLogitsHook


class BaseReducer(BaseHook, ABC):
    """
    Converts an unreduced loss tensor into a single number.
    In other words, if the loss tensor has shape ```(N,)```,
    the reducer converts it to shape ```(1,)```.
    """

    def __init__(
        self,
        apply_to: List[str] = None,
        default_reducer: "BaseReducer" = None,
        **kwargs,
    ):
        """
        Arguments:
            apply_to: list of loss names to apply reduction to
            default_reducer: a reducer to use for losses that
                are not already reduced and are also not
                specified in ```apply_to```. If ```None```,
                then no action is taken.
        """
        super().__init__(**kwargs)
        self.apply_to = apply_to
        self.default_reducer = default_reducer
        self.curr_loss_keys = []

    def call(self, losses, inputs):
        """"""
        self.curr_loss_keys = list(losses.keys())
        apply_to = self.get_keys_to_apply_to(losses)
        losses, outputs = self.call_reducer(losses, inputs, apply_to)
        if self.default_reducer:
            c_f.assert_dicts_are_disjoint(inputs, outputs)
            losses, new_outputs = self.default_reducer(losses, {**inputs, **outputs})
            outputs.update(new_outputs)
        if losses.keys() != set(self.curr_loss_keys):
            raise ValueError(
                "Loss dict returned by reducer should have same keys as input loss dict"
            )
        return losses, outputs

    @abstractmethod
    def call_reducer(self, losses, inputs, apply_to):
        pass

    def _loss_keys(self):
        """"""
        return self.curr_loss_keys

    def get_keys_to_apply_to(self, losses):
        apply_to = self.apply_to
        if apply_to is None:
            apply_to = [k for k, v in losses.items() if not c_f.len_one_tensor(v)]
        elif len(set(apply_to) - set(self.curr_loss_keys)) > 0:
            raise ValueError(
                f"self.apply_to ({self.apply_to}) must be a subset of losses.keys() ({losses.keys()})"
            )
        return apply_to

    def extra_repr(self):
        return c_f.extra_repr(self, ["apply_to"])


class MultipleReducers(BaseHook):
    def __init__(self, reducers, **kwargs):
        super().__init__(**kwargs)
        self.reducers = reducers

    def call(self, losses, inputs):
        for r in self.reducers:
            losses, inputs = r(losses, inputs)
        return losses, inputs

    def _loss_keys(self):
        return c_f.join_lists([r.loss_keys for r in self.reducers])

    def _out_keys(self):
        return c_f.join_lists([r.out_keys for r in self.reducers])


class EntropyReducer(BaseReducer):
    """
    Implementation of "entropy conditioning" from
    [Conditional Adversarial Domain Adaptation](https://arxiv.org/abs/1705.10667).
    It weights loss elements using
    [```EntropyWeights```][pytorch_adapt.layers.entropy_weights.EntropyWeights].
    The entropy weights are derived from classifier logits.
    """

    def __init__(
        self,
        f_hook: BaseHook = None,
        domains: List[str] = None,
        entropy_weights_fn: Callable[[torch.Tensor], torch.Tensor] = None,
        detach_weights: bool = True,
        **kwargs,
    ):
        """
        Arguments:
            f_hook: the hook for computing logits from
                which entropy weights are derived
            domains: the domains that ```f_hook``` should compute for
            entropy_weights_fn: the function for computing the weights
                that will be multiplied with the unreduced losses.
            detach_weights: If ```True```, the entropy weights are
                detached from the autograd graph
        """
        super().__init__(**kwargs)
        src_regex = "^{0}_|_{0}$|_{0}_|^{0}$".format("src")
        target_regex = "^{0}_|_{0}$|_{0}_|^{0}$".format("target")
        self.src_regex = re.compile(src_regex)
        self.target_regex = re.compile(target_regex)
        self.entropy_weights_fn = c_f.default(entropy_weights_fn, EntropyWeights, {})
        self.f_hook = c_f.default(
            f_hook,
            FeaturesAndLogitsHook,
            {
                "detach_features": detach_weights,
                "detach_logits": detach_weights,
                "domains": domains,
            },
        )
        self.context = torch.no_grad() if detach_weights else nullcontext()

    def call_reducer(self, losses, inputs, apply_to):
        outputs = self.f_hook(losses, inputs)[1]
        for k in apply_to:
            if self.src_regex.search(k):
                domain = "src"
            elif self.target_regex.search(k):
                domain = "target"
            else:
                raise ValueError
            with self.context:
                search_str = c_f.filter(self.f_hook.out_keys, "_logits", [f"^{domain}"])
                [logits] = c_f.extract([outputs, inputs], search_str)
                weights = self.entropy_weights_fn(logits)
            losses[k] = torch.mean(weights * losses[k])

        return losses, outputs

    def _out_keys(self):
        """"""
        return self.f_hook.out_keys


class MeanReducer(BaseReducer):
    """
    Reduces loss elements by taking the mean.
    """

    def call_reducer(self, losses, inputs, apply_to):
        for k in apply_to:
            losses[k] = torch.mean(losses[k])
        return losses, {}

    def _out_keys(self):
        """"""
        return []



import torch

from ..utils import common_functions as c_f


class ConcatSoftmax(torch.nn.Module):
    """
    Applies softmax to the concatenation of a list of tensors.
    """

    def __init__(self, dim: int = 1):
        """
        Arguments:
            dim: a dimension along which softmax will be computed
        """
        super().__init__()
        self.dim = dim

    def forward(self, *x: torch.Tensor):
        """
        Arguments:
            *x: A sequence of tensors to be concatenated
        """
        all_logits = torch.cat(x, dim=self.dim)
        return torch.nn.functional.softmax(all_logits, dim=self.dim)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["dim"])



import torch.nn as nn


class Discriminator(nn.Module):
    """
    A 3-layer MLP for domain classification.
    """

    def __init__(self, in_size=2048, h=2048, out_size=1):
        """
        Arguments:
            in_size: size of the input
            h: hidden layer size
            out_size: size of the output
        """

        super().__init__()
        self.h = h
        self.net = nn.Sequential(
            nn.Linear(in_size, h),
            nn.ReLU(),
            nn.Linear(h, h),
            nn.ReLU(),
            nn.Linear(h, out_size),
        )
        self.out_size = out_size

    def forward(self, x):
        """"""
        return self.net(x).squeeze(1)



from .base_container import BaseContainer
from .lr_schedulers import LRSchedulers
from .misc import Misc
from .models import Models
from .optimizers import Optimizers


def get_container(k):
    if k == "models":
        return Models({})
    elif k == "optimizers":
        return Optimizers({})
    elif k == "lr_schedulers":
        return LRSchedulers({})
    elif k == "misc":
        return Misc({})
    else:
        return BaseContainer({})


class MultipleContainers(BaseContainer):
    def __init__(self, **kwargs):
        self.store = kwargs

    def merge(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, BaseContainer):
                if k in self:
                    self[k].merge(v)
                else:
                    self[k] = v
            else:
                if k not in self:
                    self[k] = get_container(k)

    def create(self):
        self["models"].create()
        self["optimizers"].create_with(self["models"])
        self["lr_schedulers"].create_with(self["optimizers"])
        self["misc"].create()



import torchvision.transforms as T
from torchvision import datasets

from ..utils.constants import IMAGENET_MEAN, IMAGENET_STD
from ..utils.transforms import GrayscaleToRGB
from .combined_source_and_target import CombinedSourceAndTargetDataset
from .concat_dataset import ConcatDataset
from .mnistm import MNISTM
from .source_dataset import SourceDataset
from .target_dataset import TargetDataset


def get_multiple(domains, train, is_training, folder, dataset_getter, download):
    return ConcatDataset(
        [dataset_getter(d, train, is_training, folder, download) for d in domains]
    )


def get_datasets(
    src_domains,
    target_domains,
    folder,
    dataset_getter,
    download,
    return_target_with_labels,
):
    def getter(domains, train, is_training):
        return get_multiple(
            domains, train, is_training, folder, dataset_getter, download
        )

    output = {}
    output["src_train"] = SourceDataset(getter(src_domains, True, False))
    output["src_val"] = SourceDataset(getter(src_domains, False, False))
    output["target_train"] = TargetDataset(getter(target_domains, True, False))
    output["target_val"] = TargetDataset(getter(target_domains, False, False))
    if return_target_with_labels:
        output["target_train_with_labels"] = SourceDataset(
            getter(target_domains, True, False), domain=1
        )
        output["target_val_with_labels"] = SourceDataset(
            getter(target_domains, False, False), domain=1
        )
    output["train"] = CombinedSourceAndTargetDataset(
        SourceDataset(getter(src_domains, True, True)),
        TargetDataset(getter(target_domains, True, True)),
    )
    return output


def get_mnist(domain, train, is_training, folder, download):
    if domain == "mnist":
        transform = T.Compose(
            [
                T.Resize(32),
                T.ToTensor(),
                GrayscaleToRGB(),
                T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
            ]
        )
        return datasets.MNIST(
            folder, train=train, download=download, transform=transform
        )
    elif domain == "mnistm":
        transform = T.Compose(
            [
                T.ToTensor(),
                T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
            ]
        )
        return MNISTM(folder, train, transform, download=download)


def get_mnist_mnistm(
    src_domains, target_domains, folder, download=False, return_target_with_labels=False
):
    return get_datasets(
        src_domains,
        target_domains,
        folder,
        get_mnist,
        download,
        return_target_with_labels,
    )



from .base_weighter import BaseWeighter


class SumWeighter(BaseWeighter):
    """
    Weights the losses and then returns the **sum** of the weighted losses.
    """

    def __init__(self, **kwargs):
        super().__init__(reduction=sum, **kwargs)



import torch

from .concat_softmax import ConcatSoftmax
from .entropy_loss import EntropyLoss
from .utils import split_half


class SymNetsEntropyLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.softmax_fn = ConcatSoftmax()
        self.entropy_loss_fn = EntropyLoss(after_softmax=True)

    def forward(self, x, y):
        x = self.softmax_fn(x, y)
        x, y = split_half(x, dim=1)
        return self.entropy_loss_fn(x + y)


class SymNetsEntropyLossListInput(SymNetsEntropyLoss):
    def forward(self, x):
        return super().forward(*x)



import os
import tarfile

from torchvision.datasets.utils import download_url

from ..utils import common_functions as c_f
from .base_dataset import BaseDataset
from .utils import check_length


class MNISTM(BaseDataset):
    """
    The dataset used in "Domain-Adversarial Training of Neural Networks".
    It consists of colored MNIST digits.
    """

    url = "https://cornell.box.com/shared/static/jado7quprg6hzzdubvwzh9umr75damwi"
    filename = "mnist_m.tar.gz"
    md5 = "859df31c91afe82e80e5012ba928f279"

    def __init__(
        self, root: str, train: bool, transform=None, download=False, **kwargs
    ):
        """
        Arguments:
            root: The dataset must be located at ```<root>/mnist_m```
            train: Whether or not to use the training set.
            transform: The image transform applied to each sample.
        """
        super().__init__(domain="MNISTM", **kwargs)
        if not isinstance(train, bool):
            raise TypeError("train should be True or False")
        if download:
            try:
                self.set_paths_and_labels(root, train, transform)
            except (FileNotFoundError, ValueError):
                self.download_dataset(root)
                self.set_paths_and_labels(root, train, transform)
        else:
            self.set_paths_and_labels(root, train, transform)

    def set_paths_and_labels(self, root, train, transform):
        name = "train" if train else "test"
        labels_file = os.path.join(root, "mnist_m", f"mnist_m_{name}_labels.txt")
        img_dir = os.path.join(root, "mnist_m", f"mnist_m_{name}")
        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_length(self, {"train": 59001, "test": 9001}[name])
        self.labels = [int(x[1]) for x in content]
        self.transform = transform

    def download_dataset(self, root):
        download_url(self.url, root, filename=self.filename, md5=self.md5)
        with tarfile.open(os.path.join(root, self.filename), "r:gz") as tar:
            tar.extractall(path=root, members=c_f.extract_progress(tar))



from ..containers import KeyEnforcer
from ..hooks import MCDHook
from .base_adapter import BaseGCAdapter


class MCD(BaseGCAdapter):
    """
    Wraps [MCDHook][pytorch_adapt.hooks.mcd].
    """

    hook_cls = MCDHook

    def inference_default(self, x, domain=None):
        features = self.models["G"](x)
        logits_list = self.models["C"](features)
        logits = sum(logits_list)
        return features, logits

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            g_opts=[self.optimizers["G"]], c_opts=[self.optimizers["C"]], **hook_kwargs
        )



import torch


class BNMLoss(torch.nn.Module):
    """
    Implementation of the loss in
    [Towards Discriminability and Diversity:
    Batch Nuclear-norm Maximization
    under Label Insufficient Situations](https://arxiv.org/abs/2003.12237).
    """

    def forward(self, x):
        """"""
        x = torch.nn.functional.softmax(x, dim=1)
        return -torch.linalg.norm(x, "nuc") / x.shape[0]



from ..containers import KeyEnforcer
from ..hooks import CDANNEHook, DANNEHook, DANNHook, GVBEHook, GVBHook
from ..layers import ModelWithBridge
from .base_adapter import BaseGCDAdapter
from .gan import CDAN


class DANN(BaseGCDAdapter):
    """
    Wraps [DANNHook][pytorch_adapt.hooks.dann].
    """

    hook_cls = DANNHook

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(opts=list(self.optimizers.values()), **hook_kwargs)


class DANNE(DANN):
    hook_cls = DANNEHook


class CDANNE(DANN, CDAN):
    hook_cls = CDANNEHook


class GVB(DANN):
    """
    Wraps [GVBHook][pytorch_adapt.hooks.gvb].
    """

    hook_cls = GVBHook

    def init_containers_and_check_keys(self):
        models = self.containers["models"]
        for k in ["D", "C"]:
            if not isinstance(models[k], ModelWithBridge):
                models[k] = ModelWithBridge(models[k])
        super().init_containers_and_check_keys()


class GVBE(GVB):
    hook_cls = GVBEHook



from ..layers import EntropyLoss
from .base_validator import BaseValidator


class EntropyValidator(BaseValidator):
    """
    Returns the negative of the
    [entropy][pytorch_adapt.layers.entropy_loss.EntropyLoss]
    of all logits.
    """

    def compute_score(self, target_train):
        return -EntropyLoss()(target_train["logits"]).item()



from .base_weighter import BaseWeighter
from .mean_weighter import MeanWeighter, mean_weighter
from .sum_weighter import SumWeighter


def total_loss(loss_components):
    return loss_components["total"]


def only_components(loss_components):
    return {k: v for k, v in loss_components.items() if k != "total"}


def get_multiple_loss_totals(multiple_loss_components):
    return {k: total_loss(v) for k, v in multiple_loss_components.items()}



import torch
import torch.nn.functional as F


# https://openaccess.thecvf.com/content_iccv_2015/papers/Tzeng_Simultaneous_Deep_Transfer_ICCV_2015_paper.pdf
class UniformDistributionLoss(torch.nn.Module):
    """
    Implementation of the confusion loss from
    [Simultaneous Deep Transfer Across Domains and Tasks](https://arxiv.org/abs/1510.02192).
    """

    # *args to make it work as a drop in replacement for CrossEntropyLoss
    def forward(self, x, *args):
        """"""
        probs = F.log_softmax(x, dim=1)
        avg_probs = torch.mean(probs, dim=1)
        return -torch.mean(avg_probs)



import logging
import time

import ignite.distributed as idist
import torch
from ignite.contrib.handlers import ProgressBar
from ignite.engine import Engine, Events
from ignite.utils import setup_logger

from ...utils import common_functions as c_f
from ...utils import exceptions
from ...weighters import get_multiple_loss_totals
from .dictionary_accumulator import DictionaryAccumulator


def add_validation_runner(
    ignite_wrapper,
    dataloaders,
    validator,
    stat_getter,
    validation_interval,
    max_epochs,
    saver,
    logger,
    check_initial_accuracy,
):
    validation_condition = Events.EPOCH_COMPLETED(every=validation_interval)
    if max_epochs % validation_interval != 0:
        validation_condition |= Events.EPOCH_COMPLETED(once=max_epochs)
    if check_initial_accuracy:
        validation_condition |= Events.STARTED
    ignite_wrapper.trainer.add_event_handler(
        validation_condition,
        get_validation_runner(
            ignite_wrapper,
            dataloaders,
            validator,
            stat_getter,
            saver,
            logger,
        ),
    )


def get_validation_runner(
    cls,
    dataloaders,
    validator,
    stat_getter,
    saver,
    logger,
):
    required_data = validator.required_data
    if stat_getter:
        required_data = list(set(required_data + stat_getter.required_data))

    def run_validation(engine):

        epoch = engine.state.epoch

        collected_data = collect_from_dataloaders(cls, dataloaders, required_data)
        get_validation_score(collected_data, validator, epoch)

        if saver:
            saver.save_validator(validator)
            saver.save_adapter(cls.adapter, epoch, validator.best_epoch)
        if logger:
            logger.add_validation({"validator": validator}, epoch)
        log_str = f"VALIDATION SCORES:\n{validator}\n"

        if stat_getter:
            get_validation_score(collected_data, stat_getter, epoch)
            if saver:
                saver.save_stat_getter(stat_getter)
            if logger:
                logger.add_validation({"stat_getter": stat_getter}, epoch)
            log_str += f"OTHER STATS:\n{stat_getter}\n"

        c_f.LOGGER.info(log_str)
        if logger:
            logger.write(engine)

    return run_validation


def get_validation_score(collected_data, validator, epoch):
    return validator.score(
        epoch,
        **c_f.filter_kwargs(collected_data, validator.required_data),
    )


def collect_from_dataloaders(cls, dataloaders, required_data):
    collected_data = {}

    for k in required_data:
        c_f.LOGGER.info(f"Collecting {k}")
        curr_dataloader = dataloaders[k]
        c_f.val_dataloader_checks(curr_dataloader)
        curr_dataset = curr_dataloader.dataset
        collector = cls.get_collector(curr_dataset)
        iterable = curr_dataloader.__iter__()
        curr_collected = accumulate_collector_output(collector, iterable, k)
        c_f.val_collected_data_checks(curr_collected, curr_dataset)
        collected_data[k] = curr_collected
        del iterable

    return collected_data


def step_lr_schedulers(lr_schedulers, scheduler_type):
    def handler(engine):
        lr_schedulers.step(scheduler_type)

    return handler


def auto_model(*args, **kwargs):
    def handler(model):
        return idist.auto_model(model, *args, **kwargs)

    return handler


def save_adapter_without_validator(saver, adapter):
    def handler(engine):
        saver.save_adapter(adapter, engine.state.epoch, None)

    return handler


def early_stopper(patience, validator):
    def handler(engine):
        if engine.state.epoch > 2:
            # this runs at the beginning of a new epoch
            # so engine.state.epoch has already incremented
            # it's also 1-indexed
            epochs_since_best_epoch = engine.state.epoch - 1
            if validator.best_epoch is not None:
                epochs_since_best_epoch -= validator.best_epoch
            c_f.LOGGER.info(f"epochs_since_best_epoch = {epochs_since_best_epoch}")
            if epochs_since_best_epoch > patience:
                c_f.LOGGER.info("***Performance has plateaued. Exiting.***")
                engine.terminate()

    return handler


def pbar_print_losses(pbar):
    def handler(engine):
        losses = get_multiple_loss_totals(engine.state.output)
        losses = [f"{k}={v:6.4f}" for k, v in losses.items()]
        pbar.pbar.unit = ", ".join(losses)

    return handler


def get_default_pbar():
    bar_format = "{desc}[{n_fmt}/{total_fmt}] {percentage:3.0f}%|{bar}|{unit} [{elapsed}<{remaining}]"
    return ProgressBar(persist=True, bar_format=bar_format)


def attach_pbar(engine, *_):
    pbar = get_default_pbar()
    pbar.attach(engine)
    return pbar


def accumulate_collector_output(collector, iterable, output_name):
    accumulator = DictionaryAccumulator()
    accumulator.attach(collector, output_name)
    collector.run(iterable)
    accumulator.detach(collector)
    return collector.state.metrics[output_name]


def set_engine_logger(engine, name, level=logging.CRITICAL):
    engine.logger = setup_logger(name=name, level=level)
    return engine.logger


def register(engine, event, *args):
    for h in args:
        engine.add_event_handler(event, h)


def do_for_all_engines(cls, function):
    output = {}
    for name in ["trainer", "labeled_collector", "unlabeled_collector"]:
        output[name] = function(getattr(cls, name), name)
    return output


def resume_checks(validator, stat_getter, framework):
    last_trainer_epoch = framework.trainer.state.epoch
    for name, v in {"validator": validator, "stat_getter": stat_getter}.items():
        if not v:
            continue
        last_validator_epoch = v.epochs[-1]
        if last_trainer_epoch != last_validator_epoch:
            raise exceptions.ResumeCheckError(
                f"Last trainer epoch ({last_trainer_epoch}) does not equal last {name} epoch ({last_validator_epoch})"
            )


def is_done(trainer, max_epochs=None, **kwargs):
    try:
        return trainer.state.epoch >= max_epochs
    except TypeError:
        return False


def zero_grad(adapter):
    def handler(engine):
        c_f.LOGGER.info("zeroing grads")
        adapter.models.zero_grad()
        adapter.optimizers.zero_grad()

    return handler
